#include <gtk/gtk.h>
#include <phat/phat.h>
#include <stdlib.h>
#include "patchsection.h"
#include "specimen.h"
#include "gui.h"
#include "sampletab.h"
#include "miditab.h"
#include "voicetab.h"
#include "filtertab.h"
#include "velocitytab.h"
#include "envelopetab.h"
#include "lfotab.h"
#include "patch.h"
#include "mixer.h"

static GtkVBoxClass* parent_class;

static void patch_section_class_init(PatchSectionClass* klass);
static void patch_section_init(PatchSection* self);


typedef struct _MenuItem
{
    GtkWidget *item;
    PatchSection* self;
    int patch;
} MenuItem;


GType patch_section_get_type(void)
{
    static GType type = 0;

    if (!type)
    {
	static const GTypeInfo info =
	    {
		sizeof (PatchSectionClass),
		NULL,
		NULL,
		(GClassInitFunc) patch_section_class_init,
		NULL,
		NULL,
		sizeof (PatchSection),
		0,
		(GInstanceInitFunc) patch_section_init,
	    };

	type = g_type_register_static(GTK_TYPE_VBOX, "PatchSection", &info, 0);
    }

    return type;
}


static void patch_section_class_init(PatchSectionClass* klass)
{
    parent_class = g_type_class_peek_parent(klass);
}


static void play_cb(GtkButton* button, SampleTab* self)
{
    mixer_note_on_with_id(self->patch, patch_get_note(self->patch), 127);
}


static void stop_cb(GtkButton* button, SampleTab* self)
{
    mixer_note_off_with_id(self->patch, patch_get_note(self->patch));
}


static void vol_cb(PhatFanSlider* fan, PatchSection* self)
{
    float val;

    val = phat_fan_slider_get_value(fan);
    patch_set_volume(self->patch, val);
}


static void pan_cb(PhatFanSlider* fan, PatchSection* self)
{
    float val;

    val = phat_fan_slider_get_value(fan);
    patch_set_panning(self->patch, val);
}


static void pitch_cb(PhatFanSlider* fan, PatchSection* self)
{
    float val;

    val = phat_fan_slider_get_value(fan);
    patch_set_pitch(self->patch, val);
}


static void range_cb(PhatSliderButton* button, PatchSection* self)
{
    int val;

    val = phat_slider_button_get_value(button);
    patch_set_pitch_steps(self->patch, val);
}


static void connect(PatchSection* self)
{
    g_signal_connect(G_OBJECT(self->volume_fan), "value-changed",
		     G_CALLBACK(vol_cb), (gpointer) self);
    g_signal_connect(G_OBJECT(self->pan_fan), "value-changed",
		     G_CALLBACK(pan_cb), (gpointer) self);
    g_signal_connect(G_OBJECT(self->pitch_fan), "value-changed",
		     G_CALLBACK(pitch_cb), (gpointer) self);
    g_signal_connect(G_OBJECT(self->range_sb), "value-changed",
		     G_CALLBACK(range_cb), (gpointer) self);
    g_signal_connect(G_OBJECT(self->play_button), "pressed",
		     G_CALLBACK(play_cb), (gpointer) self);
    g_signal_connect(G_OBJECT(self->play_button), "released",
		     G_CALLBACK(stop_cb), (gpointer) self);
}


static void block(PatchSection* self)
{
    g_signal_handlers_block_by_func(self->volume_fan, vol_cb, self);
    g_signal_handlers_block_by_func(self->pan_fan, pan_cb, self);
    g_signal_handlers_block_by_func(self->pitch_fan, pitch_cb, self);
    g_signal_handlers_block_by_func(self->range_sb, range_cb, self);
}


static void unblock(PatchSection* self)
{
    g_signal_handlers_unblock_by_func(self->volume_fan, vol_cb, self);
    g_signal_handlers_unblock_by_func(self->pan_fan, pan_cb, self);
    g_signal_handlers_unblock_by_func(self->pitch_fan, pitch_cb, self);
    g_signal_handlers_unblock_by_func(self->range_sb, range_cb, self);
}


static void set_sensitive(PatchSection* self, gboolean val)
{
    gtk_widget_set_sensitive(self->patch_opt, val);
    gtk_widget_set_sensitive(self->play_button, val);
    gtk_widget_set_sensitive(self->volume_fan, val);
    gtk_widget_set_sensitive(self->pan_fan, val);
    gtk_widget_set_sensitive(self->pitch_fan, val);
    gtk_widget_set_sensitive(self->range_sb, val);
    gtk_widget_set_sensitive(self->notebook, val);
    
}


static void set_patch(PatchSection* self, int patch)
{
    float vol, pan, pitch;
    int range;

    self->patch = patch;

    if (patch < 0)
    {
	set_sensitive(self, FALSE);
    }
    else
    {
	set_sensitive(self, TRUE);

	vol = patch_get_volume(patch);
	pan = patch_get_panning(patch);
	pitch = patch_get_pitch(patch);
	range = patch_get_pitch_steps(patch);

	block(self);

	phat_fan_slider_set_value(PHAT_FAN_SLIDER(self->volume_fan), vol);
	phat_fan_slider_set_value(PHAT_FAN_SLIDER(self->pan_fan), pan);
	phat_fan_slider_set_value(PHAT_FAN_SLIDER(self->pitch_fan), pitch);
	phat_slider_button_set_value(PHAT_SLIDER_BUTTON(self->range_sb), range);
	
	unblock(self);
    }

    sample_tab_set_patch(SAMPLE_TAB(self->sample_tab), patch);
    midi_tab_set_patch(MIDI_TAB(self->midi_tab), patch);
    voice_tab_set_patch(VOICE_TAB(self->voice_tab), patch);
    filter_tab_set_patch(FILTER_TAB(self->filter_tab), patch);
    velocity_tab_set_patch(VELOCITY_TAB(self->vel_tab), patch);
    envelope_tab_set_patch(ENVELOPE_TAB(self->env_tab), patch);
    lfo_tab_set_patch(LFO_TAB(self->lfo_tab), patch);
}
	

inline static GtkWidget* play_button_new(void)
{
    GtkWidget* button;
    GtkWidget* image;

    button = gtk_button_new();
    image = gtk_image_new_from_file(PIXMAPSDIR "play.png");

    gtk_container_add(GTK_CONTAINER(button), image);
    gtk_widget_show(image);

    return button;
}


static void patch_section_init(PatchSection* self)
{
    GtkBox* box = GTK_BOX(self);
    GtkWidget* label;
    GtkWidget* table;
    GtkWidget* pad;
    GtkWidget* hbox;

    self->patch = -1;
    
    /* table */
    table = gtk_table_new(4, 7, FALSE);
    gtk_box_pack_start(box, table, TRUE, TRUE, 0);
    gtk_widget_show(table);

    /* column 2 padding */
    pad = gui_hpad_new(GUI_TEXTSPACE);
    gtk_table_attach(GTK_TABLE(table), pad, 1, 2, 0, 1, 0, 0, 0, 0);
    gtk_widget_show(pad);
    
    /* column 4 padding */
    pad = gui_hpad_new(GUI_SECSPACE);
    gtk_table_attach(GTK_TABLE(table), pad, 3, 4, 1, 2, 0, 0, 0, 0);
    gtk_widget_show(pad);

    /* column 6 padding */
    pad = gui_hpad_new(GUI_TEXTSPACE);
    gtk_table_attach(GTK_TABLE(table), pad, 5, 6, 1, 2, 0, 0, 0, 0);
    gtk_widget_show(pad);

    /* row 1 spacing */
    gtk_table_set_row_spacing(GTK_TABLE(table), 0, GUI_SPACING);
    
    /* row 3 spacing */
    gtk_table_set_row_spacing(GTK_TABLE(table), 2, GUI_SPACING*2);

    /* patch */
    label = gtk_label_new("Patch:");
    hbox = gtk_hbox_new(FALSE, GUI_SPACING);
    self->patch_opt = gtk_option_menu_new();
    self->play_button = play_button_new();

    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_box_pack_start(GTK_BOX(hbox), self->patch_opt, TRUE, TRUE, 0);
    gtk_box_pack_start(GTK_BOX(hbox), self->play_button, FALSE, FALSE, 0);
    
    gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1, GTK_FILL, 0, 0, 0);
    gtk_table_attach_defaults(GTK_TABLE(table), hbox, 2, 7, 0, 1);

    gtk_widget_show(label);
    gtk_widget_show(hbox);
    gtk_widget_show(self->patch_opt);
    gtk_widget_show(self->play_button);
        
    /* volume */
    label = gtk_label_new("Volume:");
    self->volume_fan = phat_hfan_slider_new_with_range(DEFVOL, 0.0, 1.0, 0.1);

    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2, GTK_FILL, 0, 0, 0);
    gtk_table_attach_defaults(GTK_TABLE(table), self->volume_fan, 2, 3, 1, 2);

    gtk_widget_show(label);
    gtk_widget_show(self->volume_fan);

    /* panning */
    label = gtk_label_new("Panning:");
    self->pan_fan = phat_hfan_slider_new_with_range(0.0, -1.0, 1.0, 0.1);

    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 0, 1, 2, 3, GTK_FILL, 0, 0, 0);
    gtk_table_attach_defaults(GTK_TABLE(table), self->pan_fan, 2, 3, 2, 3);

    gtk_widget_show(label);
    gtk_widget_show(self->pan_fan);

    /* pitch */
    label = gtk_label_new("Pitch:");
    self->pitch_fan = phat_hfan_slider_new_with_range(0.0, -1.0, 1.0, 0.1);

    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
    gtk_table_attach(GTK_TABLE(table), label, 4, 5, 1, 2, GTK_FILL, 0, 0, 0);
    gtk_table_attach_defaults(GTK_TABLE(table), self->pitch_fan, 6, 7, 1, 2);

    gtk_widget_show(label);
    gtk_widget_show(self->pitch_fan);

    /* range */
    label = gtk_label_new("Steps:");
    self->range_sb = phat_slider_button_new_with_range(2.0, 0.0, 48.0, 1.0, "%.0f");

    phat_slider_button_set_threshold(PHAT_SLIDER_BUTTON(self->range_sb), GUI_THRESHOLD);
    gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);

    gtk_table_attach(GTK_TABLE(table), label, 4, 5, 2, 3, GTK_FILL, 0, 0, 0);
    gtk_table_attach_defaults(GTK_TABLE(table), self->range_sb, 6, 7, 2, 3);

    gtk_widget_show(label);
    gtk_widget_show(self->range_sb);

    /* notebook */
    self->notebook = gtk_notebook_new();
    gtk_table_attach_defaults(GTK_TABLE(table), self->notebook, 0, 7, 3, 4);
    gtk_widget_show(self->notebook);

    /* sample page */
    self->sample_tab = sample_tab_new();
    label = gtk_label_new("SMP");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->sample_tab, label);
    gtk_widget_show(self->sample_tab);
    gtk_widget_show(label);
    
    /* midi page */
    self->midi_tab = midi_tab_new();
    label = gtk_label_new("MIDI");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->midi_tab, label);
    gtk_widget_show(self->midi_tab);
    gtk_widget_show(label);

    /* voice page */
    self->voice_tab = voice_tab_new();
    label = gtk_label_new("VOICE");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->voice_tab, label);
    gtk_widget_show(self->voice_tab);
    gtk_widget_show(label);

    /* filter page */
    self->filter_tab = filter_tab_new();
    label = gtk_label_new("FILT");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->filter_tab, label);
    gtk_widget_show(self->filter_tab);
    gtk_widget_show(label);

    /* velocity page */
    self->vel_tab = velocity_tab_new();
    label = gtk_label_new("VEL");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->vel_tab, label);
    gtk_widget_show(self->vel_tab);
    gtk_widget_show(label);

    /* envelope page */
    self->env_tab = envelope_tab_new();
    label = gtk_label_new("ENV");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->env_tab, label);
    gtk_widget_show(self->env_tab);
    gtk_widget_show(label);

    /* lfo page */
    self->lfo_tab = lfo_tab_new();
    label = gtk_label_new("LFO");
    gtk_notebook_append_page(GTK_NOTEBOOK(self->notebook), self->lfo_tab, label);
    gtk_widget_show(self->lfo_tab);
    gtk_widget_show(label);

    /* build connections */
    connect(self);
}


GtkWidget* patch_section_new(void)
{
    return (GtkWidget*) g_object_new(PATCH_SECTION_TYPE, NULL);
}


static void menu_item_cb(GtkWidget* widget, gpointer data)
{
    MenuItem* item = (MenuItem*)data;
    
    set_patch(item->self, item->patch);
}


void patch_section_update(PatchSection* self, int index)
{
    static int count = 0;
    static GtkWidget *menu = NULL;
    static MenuItem* items = NULL;
    GtkOptionMenu* opt = GTK_OPTION_MENU(self->patch_opt);
    int *patch;			/* array of active patches gets dumped here */
    int i;
    char *name;

    /* destroy any menu items */
    if (items != NULL)
    {
	for (i = 0; i < count; i++)
	{
	    gtk_widget_destroy(items[i].item);
	}

	items = NULL;
    }

    /* destroy menu */
    if (menu != NULL)
	gtk_widget_destroy(menu);

    /* create new menu */
    menu = gtk_menu_new();

    /* get array of active patches */
    count = patch_dump(&patch);
    if (count < 0)
	debug("Failed to update interface, couldn't get array of active patches!\n");

    if (count == 0)
    {
	/* better to attach an empty menu than no menu at all */
	gtk_option_menu_set_menu(opt, menu);

	set_patch(self, -1);

	return;
    }

    /* create array of menu items */
    items = g_new(MenuItem, count);

    /* create a menu item for each patch id with appropriate callback */
    for (i = 0; i < count; i++)
    {
	if ((name = patch_get_name(patch[i])) == NULL)
	{
	    errmsg("Couldn't find name for patch %d.\n", patch[i]);
	    exit (FUBAR);
	}

	items[i].item = gtk_menu_item_new_with_label(name);
	items[i].patch = patch[i];
	items[i].self = self;
	
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), items[i].item);

	g_signal_connect(G_OBJECT(items[i].item), "activate",
			  G_CALLBACK(menu_item_cb),
			  (gpointer)&items[i]);

	gtk_widget_show(items[i].item);

	g_free(name);
    }

    /* free patch array */
    g_free(patch);

    /* attach new menu to patch option-menu */
    gtk_option_menu_set_menu(GTK_OPTION_MENU(opt), menu);

    /* turn to the requested item */
    gtk_menu_item_activate(GTK_MENU_ITEM(items[index].item));
    gtk_option_menu_set_history(opt, index);
}


int patch_section_get_current_patch(PatchSection* self)
{
    return self->patch;
}
